
public class Solver
	implements Runnable
{
	final private TwisterCanvas parent;
	private boolean abort = false;
	public void stop(){ abort = true; }

	public Solver( TwisterCanvas c ){
		parent = c;
		init();
	}

	// 0..8=normal, 9-17=flipped
	final private int[][] tran = {
			{ 1,4,2,0,3,5,6,7,8,  10,13,11,9,12,14,15,16,17 },//1 ul
			{ 0,11,14,3,10,13,6,7,8,  9,2,5,12,1,4,15,16,17 },//2 ur
			{ 0,1,2,3,5,8,6,4,7,  9,10,11,12,14,17,15,13,16 },//5 dr
			{ 0,1,2,13,16,5,12,15 ,8,  9,10,11,4,7,14,3,6,17 },//4 dl
	};
	final private int PRUNLEN = 18*18*18;
	final private int[] prun = new int[PRUNLEN];

	private void calcPrun(){
		for( int i=0; i<PRUNLEN; i++) prun[i] = -1;
		prun[(0*18+1)*18+2] = 0; // tiles 123 solved.
		int len = 0;
		int n=1;
		while( n>0 && !abort){
			n=0;
			for( int i=0; i<PRUNLEN && !abort; i++){
				if( prun[i] == len ){
					int p = i;
					int p3 = p%18; p/=18;
					int p2 = p%18; p/=18;
					int p1 = p;
					for( int m=0; m<4; m++ ){
						for( int q=0; q<4; q++ ){
							p1 = tran[m][p1];
							p2 = tran[m][p2];
							p3 = tran[m][p3];
							p = (p1*18+p2)*18+p3;
							if( prun[p] < 0 ){
								prun[p] = len+1;
								n++;
							}
						}
					}
				}
			}
			len++;
		}
	}

	private void applyTran( int[] idx, int m, int q ){
		idx[0] = applyTran( idx[0], m, q ); // top
		if( m==1 || m==3 ) m ^= 2;
		q = (-q)&3;
		idx[1] = applyTran( idx[1], m, q ); // left
		if( m==0 || m==2 ) m ^= 2;
		q = (-q)&3;
		idx[2] = applyTran( idx[2], m, q ); // bottom
		if( m==1 || m==3 ) m ^= 2;
		q = (-q)&3;
		idx[3] = applyTran( idx[3], m, q ); // right
	}

	// apply move m, q times
	private int applyTran( int p, int m, int q ){
		int p3 = p%18; p/=18;
		int p2 = p%18; p/=18;
		int p1 = p;
		while( q>0 ){
			p1 = tran[m][p1];
			p2 = tran[m][p2];
			p3 = tran[m][p3];
			q--;
		}
		p = (p1*18+p2)*18+p3;
		return p;
	}


	// oricalc = quarter turns to get it to top left pos without changing flip
	private final int[] oricalc = { 0, 3, 0, 1, 2, 1, 0, 3, 0 };
	private final int[] sym1 = { 0,3,6,1,4,7,2,5,8,  9,12,15,10,13,16,11,14,17 }; // left edge
	private final int[] sym2 = { 8,7,6,5,4,3,2,1,0,  17,16,15,14,13,12,11,10,9 }; // bottom edge
	private final int[] sym3 = { 8,5,2,7,4,1,6,3,0,  17,14,11,16,13,10,15,12,9 }; // right edge

	private int[] dosolve( int[] pos, int[] ori ){
		// create list of piece positions, 0-8=normal, 9-17=flipped
		int[] pos2 = new int[9];
		for( int i=0; i<9; i++ ){
			int pc = pos[i];
			int pcor = (ori[i] + oricalc[i] - oricalc[pc] )&3;
			pos2[pc] = pcor!=0 ? i + 9 : i;
		}

		// create list of edge triplet positions
		int[] pos3 = new int[4];
		pos3[0] = (      pos2[0] *18 +      pos2[1] )*18 +      pos2[2] ; // top
		pos3[1] = ( sym1[pos2[0]]*18 + sym1[pos2[3]])*18 + sym1[pos2[6]]; // left
		pos3[2] = ( sym2[pos2[8]]*18 + sym2[pos2[7]])*18 + sym2[pos2[6]]; // bottom
		pos3[3] = ( sym3[pos2[8]]*18 + sym3[pos2[5]])*18 + sym3[pos2[2]]; // right

		// find solution
		int[] buf = new int[10];
		int len = 0;
		while( !findsol( pos3, -1, 0, len, buf ) && len<=buf.length && !abort ){
			len++;
		}
		if( len>buf.length ) return null;
		int[] sol = new int[len];
		for( int i=0; i<len; i++ ) sol[i] = buf[i];
		return sol;
	}

	private boolean findsol( int[] idx, int lm, int dpth, int rem, int[] sol ){
		if( abort ) return false;
		// test if still solvable
		for( int i=0; i<4; i++ ){
			if( prun[idx[i]] > rem ) return false;
		}
		// test if finished.
		if( rem==0 ) return true;
		if( dpth==sol.length ) return true;

		// try every move
		for( int m=0; m<4; m++ ){
			if( m!=lm ){
				applyTran( idx, m, 1 );
				for( int q=1; q<4; q++ ){
					sol[dpth] = m*4+q;
					boolean found = findsol( idx, m, dpth+1, rem-1, sol );
					if( found ) return true;
					applyTran( idx, m, 1 );
				}
			}
		}
		return false;
	}

	private final int IDLE_MODE = 0; 
	private final int INIT_MODE = 1; 
	private final int SOLVE_MODE = 2; 
	private int mode = IDLE_MODE;
	private int[] savedPos, savedOri;
	public void run() {
		if( mode == INIT_MODE ){
			calcPrun();
		}else if( mode == SOLVE_MODE ){
			int[] sol = dosolve( savedPos, savedOri );
			parent.solverFinished( sol );
		}
		mode = IDLE_MODE;
	}
	
	private void init(){
		if( mode==IDLE_MODE ){
			mode = INIT_MODE;
			new Thread(this).start();
		}
	}
	public void solve( int[] pos, int[] ori ){
		if( mode==IDLE_MODE ){
			mode = SOLVE_MODE;
			savedPos = pos;
			savedOri = ori;
			new Thread(this).start();		
		}
	}
}
