import javax.microedition.lcdui.Display;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;
import javax.microedition.rms.RecordStore;

public class Twister extends MIDlet {
    private MenuList menuList;
    private SettingsForm settingsForm;
    private boolean initDone = false;
    private TwisterCanvas twisterCanvas;

	public Twister() {
	}

	protected void destroyApp(boolean arg0){
		twisterCanvas.stop();
	}

	protected void pauseApp() {
		twisterCanvas.pause();
	}

	protected void startApp() throws MIDletStateChangeException {
        init();   // if not already done
        Display.getDisplay(this).setCurrent(menuList);
	}

    private synchronized void init()
    {
        if (!initDone)
        {
            menuList = new MenuList(this);
            twisterCanvas = new TwisterCanvas(this);
            settingsForm = new SettingsForm( this );
            readStoreValues();
        }
    }


	public void twisterCanvasMenu( boolean solved ){
		twisterCanvas.pause();
        if( solved ) menuList.setGameRetry();
        else menuList.setGameContinue();
        Display.getDisplay(this).setCurrent(menuList);
    }

	public void settingsFormExit(){
    	int af = settingsForm.getAnimFrames();
    	int sv = settingsForm.getSolvingSpeed();
    	int sf = settingsForm.getShufflingSpeed();
    	if( af != twisterCanvas.getAnimFrames() ||
    		sv != twisterCanvas.getSolvingSpeed() ||
    		sf != twisterCanvas.getShufflingSpeed() ){
    		writeStoreValues(af, sv, sf);
    	}
        Display.getDisplay(this).setCurrent(menuList);
    }

    public void menuListContinue(){
        Display.getDisplay(this).setCurrent(twisterCanvas);
		twisterCanvas.cont();
    }
    public void menuListRetry(){
		twisterCanvas.restartGame();
        Display.getDisplay(this).setCurrent(twisterCanvas);
		twisterCanvas.cont();
    }
    public void menuListSolve(){
        Display.getDisplay(this).setCurrent(twisterCanvas);
		twisterCanvas.cont();
		twisterCanvas.solve();
    }
    public void menuListNewGame()
    {
    	twisterCanvas.newGame();
        Display.getDisplay(this).setCurrent(twisterCanvas);
		twisterCanvas.cont();
    }
    public void menuListInstructions()
    {
        // create and discard a new Instructions screen each time, to
        // avoid keeping heap memory for it when it's not in use
        Display.getDisplay(this).setCurrent(new InstructionsScreen(this));
    }
    public void menuListSettings(){
   		int animFrames = twisterCanvas.getAnimFrames();
   		int shufflingSpeed = twisterCanvas.getShufflingSpeed();
   		int solvingSpeed = twisterCanvas.getSolvingSpeed();
   		settingsForm.setValues( animFrames, shufflingSpeed, solvingSpeed );
   		Display.getDisplay(this).setCurrent(settingsForm);
    }
    public void menuListQuit()
    {
        quit();
    }


    public void instructionsBack()
    {
        Display.getDisplay(this).setCurrent(menuList);
    }


    private void quit()
    {
        destroyApp(false);
        notifyDestroyed();
    }


    private final String rsname = "Twister";
    private void readStoreValues(){
    	int af = 10, sv=3, sf = 5;
        RecordStore rs = null;
        try{
            rs = RecordStore.openRecordStore(rsname, false);
            if( rs!=null ){
            	byte[] data = rs.getRecord(1);
            	if( data!=null  ){
            		af = data[0];
            		sv = data[1];
            		sf = data[2];
            	}
            }
        }
        catch (Exception ex){}
        finally{
            if (rs != null){
                try{ rs.closeRecordStore(); }
                catch (Exception ex){}
            }
        }
		twisterCanvas.setAnimFrames( af );
		twisterCanvas.setShufflingSpeed( sf );
		twisterCanvas.setSolvingSpeed( sv );
    }
    private void writeStoreValues( int af, int sv, int sf ){
        RecordStore rs = null;
        try{
            rs = RecordStore.openRecordStore( rsname, true );
            byte[] data = { (byte)af, (byte)sv, (byte)sf };
            if (rs.getNumRecords() == 0 ){
            	// new record store
                rs.addRecord(data, 0, data.length);
            }else{
                // existing record store: will have one record, id 1
                rs.setRecord(1, data, 0, data.length);
            }
        }
        catch ( Exception ex ){}
        finally {
            if (rs != null){
                try{ rs.closeRecordStore(); }
                catch(Exception ex){}
            }
        }
		twisterCanvas.setAnimFrames( af );
		twisterCanvas.setShufflingSpeed( sf );
		twisterCanvas.setSolvingSpeed( sv );
    }
}