import java.util.Random;

import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.GameCanvas;
import javax.microedition.lcdui.game.Sprite;


public class TwisterCanvas extends GameCanvas
	implements Runnable, CommandListener
{
	private int shufflingSpeed = 7;
	private int solvingSpeed = 3;
	private int finaleSpeed = 9;
	private int animatingSpeed = 9;
	private int pollingSpeed = 7;
	private int animNumStep = 5; // number of frames in a move animation


	private int mode = 0;
	final private int NORMAL_MODE = 0;
	final private int SHUFFLE_MODE = 1;
	final private int FINALE_MODE = 2;
	final private int USERFINALE_MODE = 3;
	final private int SEARCHING_MODE = 4;
	final private int SOLVING_MODE = 5;

	private boolean animating = false;
	private int animAngle;	// angle at this moment during animation of a move 
	private int animAngleStart; // starting angle of move (ends at 0)
	private int animStep; // current animation frame
	private int animx, animy; //center of rotation of a move
	private int animMask;	// bitmap for which tile locations are moving

	private int lastMove = -1;
	private int moveCount = 0;
	private final Command backCommand;

	final private Twister midlet;
	final private Random random = new Random();
	final private Queue movestodo = new Queue();
	final private Image[] tiles = new Image[9];
	final private Solver solver;
	
	private int tileSize, offsetx, offsety, width, height;
	private int messagey, messageh;
	// 16
	//1  2
	// 32
	//4  8
	// 64
	
	final private int numbers[]={10, 118, 122, 43, 121, 125, 26, 127, 123 };
	final private int rotate[] = {Sprite.TRANS_NONE, Sprite.TRANS_ROT90, Sprite.TRANS_ROT180, Sprite.TRANS_ROT270 };
	final private int moves[][] = { {0,1,4,3}, {1,2,5,4}, {4,5,8,7}, {3,4,7,6} };

	final private int pos[] = new int[9];
	final private int ori[] = new int[9];
	final private int initpos[] = new int[9];
	final private int initori[] = new int[9];

	protected TwisterCanvas( Twister midlet ) {
		super(false);
		this.midlet = midlet;

		// calculate screen dimensions
		Graphics g = getGraphics();
		messageh = g.getFont().getHeight();
	    width = getWidth();
	    height = getHeight() - messageh;
	    messagey = height;
	    tileSize = ( ( width>height ? height : width ) - 10 ) / 3;
	    offsetx = (width-3*tileSize)/2;
	    offsety = (height-3*tileSize)/2;

	    // build tile images
	    Image mutable  = Image.createImage( tileSize, tileSize);
	    g = mutable.getGraphics();
		for( int d = 0; d<9; d++ ){
			drawTile2(g, d, 0, 0, 0,  0,0,0);
		    tiles[d] = Image.createImage(mutable);
		}

		backCommand = new Command("Menu", Command.BACK, 2);
	    addCommand(backCommand);
	    setCommandListener(this);

	    // reset position
		for( int i=0; i<9; i++ ){
			pos[i] = i;
			ori[i] = 0;
		}
		
		// start animation thread
		new Thread( this ).start();
		// set up solver
		solver = new Solver(this);
	}

	public void newGame(){
		if( mode == SHUFFLE_MODE ){
			synchronized( movestodo ){
				movestodo.clear();
			}
			mode = NORMAL_MODE;
		}
		mode = NORMAL_MODE;
		for( int i=0; i<9; i++ ){
			pos[i] = i;
			ori[i] = 0;
		}
		lastMove = -1;
		moveCount = 0;
		shuffle();
	}

    public void keyPressed(int keyCode)
    {
        // The constructor suppresses key events for game keys, so we'll
        // only get key events for non-game keys. The number keys, * & #
        // have positive keyCodes, so negative keyCodes mean non-game
        // special keys like soft-keys. We'll use key-presses on special
        // keys to take us to the menu.
    	if( keyCode == KEY_NUM1 ){
    		performMove( 0 );
    	} else if( keyCode == KEY_NUM2 || keyCode == KEY_NUM3){
    		performMove( 1 );
    	} else if( keyCode == KEY_NUM4 || keyCode == KEY_NUM7){
    		performMove( 3 );
    	} else if( keyCode == KEY_NUM5 || keyCode == KEY_NUM9){
    		performMove( 2 );
        }
    }
	public void commandAction(Command c, Displayable d) {
		if( c==backCommand ){
			if( mode == SEARCHING_MODE ){
				solver.stop();
				mode = NORMAL_MODE;
			}
			midlet.twisterCanvasMenu( mode == SOLVING_MODE );
		}
	}
    private void performMove( int m ){
		if( mode == NORMAL_MODE ){
			if( m!=lastMove ){
				moveCount++;
				lastMove = m;
			}
			pushmove( m, 1);
		}
    }
    private boolean isSolved(){
		for( int i=0; i<9; i++ ){
			if( pos[i]!=i || ori[i]!=0 ) return false;
		}
		return true;
    }
    
    public void solve(){
    	if( mode == NORMAL_MODE ){
        	mode = SEARCHING_MODE;
    		solver.solve( pos, ori );
    		repaint();
    	}
    }
    public void solverFinished( int[] sol ){
    	if( sol !=null && sol.length>0 ){
        	mode = SOLVING_MODE;
			synchronized( movestodo ){
				for( int i=0; i<sol.length; i++ ){
					pushmove( sol[i]>>2, sol[i]&3 );
				}
			}
    	}else{
        	mode = NORMAL_MODE;    		
    	}
    }
    
    public void paint( Graphics g ){
    	// blank message area
    	g.setColor( 0x80ff80 );
    	g.fillRect( 0, messagey, width, messageh );

    	// write message
		g.setColor( 0x0000ff );
		g.drawString( ""+moveCount, 0, messagey, Graphics.TOP|Graphics.LEFT );
    	if( mode == SHUFFLE_MODE ){
    		g.drawString( "Shuffling... ("+movestodo.size()+")", width/2, messagey, Graphics.HCENTER|Graphics.TOP );
    	}else if( mode == USERFINALE_MODE ){
    		g.drawString( "Congratulations!", width/2, messagey, Graphics.HCENTER|Graphics.TOP );
    	}else if( mode == SEARCHING_MODE ){
    		g.drawString( "Searching...", width/2, messagey, Graphics.HCENTER|Graphics.TOP );
    	}else if( mode == SOLVING_MODE ){
    		g.drawString( "Solving... ("+movestodo.size()+")", width/2, messagey, Graphics.HCENTER|Graphics.TOP );
    	}
    		

		// draw tiles
		g.setColor( 0xffffff );
		g.fillRect( 0, 0, width, height );
		if( mode == FINALE_MODE || mode == USERFINALE_MODE ){
    		for( int c=0; c<9; c++ ){
    			int x = finaleCoords[c][0]>>8;
    			int y = finaleCoords[c][1]>>8;
    			if( finaleCoords[c][2]!=0 )
    				drawTile2(g, c, x, y, 0, x + tileSize/2, y + tileSize/2, finaleCoords[c][2] );
    			else
					g.drawRegion( tiles[c], 0, 0, tileSize, tileSize, rotate[0], x, y, 0 );
    		}
    	}else if( !animating ){
    		synchronized( this ){
        		int c=0;
    			for( int i=0; i<3; i++ ){
    				for( int j=0; j<3; j++ ){
    					g.drawRegion( tiles[pos[c]], 0, 0, tileSize, tileSize, rotate[ori[c]], offsetx + j*tileSize, offsety + i*tileSize, 0 );
    					c++;
    				}
    			}
    		}
    	}else{
    		synchronized( this ){
    			// first draw all stationary tiles
    			int m = animMask;
    			int c=0;
    			for( int i=0; i<3; i++ ){
    				for( int j=0; j<3; j++ ){
    					if( (m&1)==0 ) 
    						g.drawRegion( tiles[pos[c]], 0, 0, tileSize, tileSize, rotate[ori[c]], offsetx + j*tileSize, offsety + i*tileSize, 0 );
    					m>>=1;
    					c++;
    				}
    			}
    			// now draw all moving tiles
    			m = animMask;
    			c=0;
    			for( int i=0; i<3; i++ ){
    				for( int j=0; j<3; j++ ){
    					if( (m&1)!=0 ){ 
    						drawTile2(g, pos[c], offsetx + j*tileSize, offsety + i*tileSize, ori[c], animx, animy, animAngle );
    					}
    					m>>=1;
    					c++;
    				}
    			}
    		}
    	}
    }
    
    private void animMove( int move, int q ){
		synchronized( this ){
			domove(move, q);
			animStep = 0;
			if( animNumStep == 0 ){
				animAngleStart = 0;
				animMask = 0;
			}else{
				if(q>=2) q-=4;
				animAngleStart = -q*90*256;
				if( move==0 ){
					animx = offsetx + tileSize;
					animy = offsety + tileSize;
					animMask = 1+2+8+16;
				}else if( move==1 ){
					animx = offsetx + tileSize*2;
					animy = offsety + tileSize;
					animMask = 2+4+16+32;
				}else if( move==2 ){
					animx = offsetx + tileSize*2;
					animy = offsety + tileSize*2;
					animMask = 16+32+128+256;
				}else if( move==3 ){
					animx = offsetx + tileSize;
					animy = offsety + tileSize*2;
					animMask = 8+16+64+128;
				}
			}
			animAngle = animAngleStart;
			animating = true;
		}
    }
    
    private void domove( int move, int q ){
		synchronized( this ){
			int[] m = moves[move];
			while( q>0 ){
				int p3 = pos[m[3]];
				int o3 = ori[m[3]];
				for( int i=3; i>0; i-- ){
					pos[m[i]] = pos[m[i-1]];
					ori[m[i]] = ( ori[m[i-1]] + 1 )&3;
				}
				pos[m[0]] = p3;
				ori[m[0]] = ( o3 + 1 )&3;
				q--;
			}
		}
    }

    private void shuffle(){
    	if( mode == NORMAL_MODE ){
			synchronized( movestodo ){
				movestodo.clear();
				lastMove = -1;
				moveCount = 0;
				int lm = -1;
				for( int i=0; i<30; i++ ){
					int m = random.nextInt(4);
					while( m == lm ) m = random.nextInt(4);
					lm = m;
					int q = 1 + random.nextInt(3);
					if( shufflingSpeed < 10 ) pushmove( m, q );
					else domove( m, q );
				}
				if( shufflingSpeed < 10 ){ 
					mode = SHUFFLE_MODE;
				}else{
					startGame();
				}
			}
			repaint();
    	}
    }
    
    private boolean paused = false;
    private boolean stopped = false;
    public void pause(){ paused=true; }
    public void cont(){ paused=false; }
    public void stop(){ stopped=true; solver.stop(); }
	public void run() {
		while( !stopped ){
			int speed = pollingSpeed;
			if( paused ){
			}else if( animating ){
				speed = animatingSpeed;
			}else if( mode == SHUFFLE_MODE ){
				speed = shufflingSpeed;
			}else if( mode == SOLVING_MODE ){
				speed = solvingSpeed;
			}else if( mode == FINALE_MODE || mode == USERFINALE_MODE ){
				speed = finaleSpeed;
			}else if( mode == NORMAL_MODE ){
				speed = movestodo.isEmpty() ? pollingSpeed : 10;
			}
			try{
				Thread.sleep( (10-speed)*(10-speed)*10 );
			}catch( Exception e ){
				break;
			}
			if( paused ){
			}else if( animating ){
				
				doAnimationStep();
			}else if( !movestodo.isEmpty() ){
				synchronized( movestodo ){
					popmove();
				}
			}else if( mode == FINALE_MODE || mode == USERFINALE_MODE ){
				finaleMove();
			}
		}
	}
	private void popmove(){
		if( !movestodo.isEmpty() ){
			int m = movestodo.pop();
			int q = m&3;
			m >>= 2;
			animMove(m,q);
			repaint();
		}
	}
	private void pushmove(int m, int q){
		movestodo.add(m*4+q);
	}
	private void startGame(){
		for( int i=0; i<9; i++ ){
			initpos[i] = pos[i];
			initori[i] = ori[i];
		}
		lastMove = -1;
		moveCount = 0;
		mode = NORMAL_MODE;
	}
	public void restartGame(){
		if( mode == NORMAL_MODE ){
			synchronized( this ){
				for( int i=0; i<9; i++ ){
					pos[i] = initpos[i];
					ori[i] = initori[i];
				}
				lastMove = -1;
				moveCount = 0;
			}
			repaint();
		}
	}

	private int[][] finaleCoords = new int[9][6];
	private void finaleStart( boolean user ){
    	int c=0;
    	for( int i=0; i<3; i++ ){
        	for( int j=0; j<3; j++ ){
        		finaleCoords[c][0] = (offsetx + j*tileSize)<<8;
        		finaleCoords[c][1] = (offsety + i*tileSize)<<8;
        		finaleCoords[c][2] = 0;
        		finaleCoords[c][3] = (j-1)*100 + random.nextInt(100)-50;
        		finaleCoords[c][4] = (i-3)*100 - random.nextInt(50);
        		finaleCoords[c][5] = animNumStep==0 ? 0: (j-1)*100  + random.nextInt(100)-50;
        		c++;
        	}
    	}
    	mode = user ? USERFINALE_MODE: FINALE_MODE;
	}
	private void finaleMove(){
		boolean ended = true;
		for( int i=0; i<9; i++ ){
    		finaleCoords[i][0] += finaleCoords[i][3];
    		finaleCoords[i][1] += finaleCoords[i][4];
    		finaleCoords[i][2] += finaleCoords[i][5];
    		finaleCoords[i][4] += 30;
    		if( ( finaleCoords[i][1]>>8 ) < 2*height ) ended = false;
		}
		if( ended ) {
	    	mode = NORMAL_MODE;
            midlet.twisterCanvasMenu( true );
		}
		repaint();
	}

	private void doAnimationStep(){
		synchronized( this ){
			animStep++;
			if( animStep >= animNumStep  ){
				animating = false;
				if( mode == NORMAL_MODE && isSolved() ){
					finaleStart( true );
				}else if( mode == SHUFFLE_MODE && movestodo.isEmpty() ){
					startGame();
				}else if( mode == SOLVING_MODE && movestodo.isEmpty() ){
					finaleStart( false );
				}
			}else{
				animAngle = animAngleStart * ( animNumStep - animStep )/ animNumStep;
			}
		}
		repaint();
	}
	
	public void setShufflingSpeed( int s ){ shufflingSpeed = s; }
	public void setSolvingSpeed( int s ){ solvingSpeed = s; }
	public int getShufflingSpeed(){ return shufflingSpeed; }
	public int getSolvingSpeed(){ return solvingSpeed; }
	public void setAnimFrames( int s ){ animNumStep = s; }
	public int getAnimFrames(){ return animNumStep; }

	private void drawTile2( Graphics g, int d, int x, int y, int ori, int cx, int cy, int angle){
	    g.setColor( 0xe0e0e0 );
	    Vector2D.fillRect(g, x, y, tileSize-1, tileSize-1, cx, cy, angle);
		g.setColor( 0 );
		Vector2D.drawRect(g, x, y, tileSize-1, tileSize-1, cx, cy, angle );

	    int pix = tileSize/11;
	    int offy = (tileSize - 9*pix )/2;
	    int offx = (tileSize - 5*pix )/2;
	    
	    g.setColor( 0xd0d0d0 );
	    for( int i=0; i<7; i++ )
	    	drawStroke2( g, i, pix, x, y, offx, offy, ori, cx,cy,angle );
		g.setColor( 0 );
	    for( int i=0; i<7; i++ ){
	    	if( (numbers[d] & (1<<i) )!=0 )
	    		drawStroke2( g, i, pix, x, y, offx, offy, ori, cx,cy,angle );
	    }
	    drawStroke2( g, 7, pix, x, y, offx, offy, ori, cx,cy,angle );
	}
	final private int strokes[]={0,0,0,4, 4,0,0,4, 0,4,0,4, 4,4,0,4, 0,0,4,0, 0,4,4,0, 0,8,4,0, 6,8,0,0 };
	private void drawStroke2( Graphics g, int stroke, int pixsz, int x, int y, int offx, int offy, int ori, int cx, int cy, int angle){
		int ox = strokes[stroke*4  ];
		int oy = strokes[stroke*4+1];
		int dx = (strokes[stroke*4+2]+1)*pixsz;
		int dy = (strokes[stroke*4+3]+1)*pixsz;
		int w=0 , h=0;
		if( ori == 0 ){
			x +=  offx +pixsz*ox;
			y +=  offy +pixsz*oy;
			w =  dx;
			h =  dy;
		}else if( ori == 1 ){
			x +=  tileSize - offy - pixsz*oy -dy+1;
			y +=  offx + pixsz*ox;
			w = dy;
			h = dx;
		}else if( ori == 2 ){
			x +=  tileSize - offx - pixsz*ox -dx+1;
			y +=  tileSize - offy - pixsz*oy -dy+1;
			w = dx;
			h = dy;
		}else if( ori == 3 ){
			x +=  offy + pixsz*oy;
			y +=  tileSize - offx - pixsz*ox -dx+1;
			w = dy;
			h = dx;
		}
		
		Vector2D.fillRect( g, x, y, w, h, cx, cy, angle );
	}
}
