import javax.microedition.lcdui.Graphics;


public class Vector2D {
	static final int TABLESIZE = 90;	// integer used as a right angle
	static final int FIXED_POINT = 8;	// Number of bits used for fractional part.
	static final int RIGHT_ANGLE = TABLESIZE<<FIXED_POINT;  // Fixed point value used as a right angle
	static final int FRAC_MASK = (1<<FIXED_POINT)-1;  // Bitmask used to get fractional part
	static int[] cos = new int[TABLESIZE+1];
	/*
	   cosx = 1 - x^2/2! ...
	   Approximate it by a quadratic.
	 */
	static {
		for( int a = 0; a<=TABLESIZE; a++ ){
			cos[a] = (((TABLESIZE - a) * (TABLESIZE + a))<<FIXED_POINT) /(TABLESIZE*TABLESIZE);
		}
	}
	
	static public int cos( int angle ){
		if( angle < 0 ) angle = -angle;
		do{
			if( angle < RIGHT_ANGLE ) return cosLookup( angle );
			angle -= RIGHT_ANGLE;
			if( angle < RIGHT_ANGLE ) return -cosLookup( RIGHT_ANGLE-angle );
			angle -= RIGHT_ANGLE;
			if( angle < RIGHT_ANGLE ) return -cosLookup( angle );
			angle -= RIGHT_ANGLE;
			if( angle < RIGHT_ANGLE ) return cosLookup( RIGHT_ANGLE-angle );
			angle -= RIGHT_ANGLE;
		}while(true);
	}
	static public int cosLookup( int angle ){
		int frac = angle & FRAC_MASK;
		int a = angle >> FIXED_POINT;
		int v1 = cos[a];
		if( frac==0 ) return v1;
		int v2 = cos[a+1];
		return ( v1*( (1<<FIXED_POINT)-frac) + v2*frac ) >> FIXED_POINT;
	}
	static public int sin( int angle ){
		return cos( RIGHT_ANGLE - angle );
	}


	// draws rectangle at x, y of size w,h, but rotated by angle about centre cx,cy
	static public void drawRect( Graphics g, int x, int y, int w, int h, int cx, int cy, int angle ){
		if( angle==0 ){
			g.drawRect( x, y, w, h );
		}else{
			x-=cx;
			y-=cy;
			int c = cos( angle );
			int s = sin( angle );
			int x1 = cx + ( ( x   *c- y   *s)>>FIXED_POINT );
			int y1 = cy + ( ( x   *s+ y   *c)>>FIXED_POINT );
			int x2 = cx + ( ( x   *c-(y+h)*s)>>FIXED_POINT );
			int y2 = cy + ( ( x   *s+(y+h)*c)>>FIXED_POINT );
			int x3 = cx + ( ((x+w)*c-(y+h)*s)>>FIXED_POINT );
			int y3 = cy + ( ((x+w)*s+(y+h)*c)>>FIXED_POINT );
			int x4 = cx + ( ((x+w)*c- y   *s)>>FIXED_POINT );
			int y4 = cy + ( ((x+w)*s+ y   *c)>>FIXED_POINT );
			g.drawLine(x1,y1, x2,y2);
			g.drawLine(x2,y2, x3,y3);
			g.drawLine(x3,y3, x4,y4);
			g.drawLine(x4,y4, x1,y1);
		}
	}
	// fills rectangle at x, y of size w,h, but rotated by angle about centre cx,cy
	static public void fillRect( Graphics g, int x, int y, int w, int h, int cx, int cy, int angle ){
		if( angle==0 ){
			g.fillRect( x, y, w, h );
		}else{
			x-=cx;
			y-=cy;
			int c = cos( angle );
			int s = sin( angle );
			int c2 = c>>1;
			int s2 = s>>1;
			int x1 = cx + ( ( x   *c   - y   *s   )>>FIXED_POINT );
			int y1 = cy + ( ( x   *s   + y   *c   )>>FIXED_POINT );
			int x2 = cx + ( ( x   *c   -(y+h)*s+s2)>>FIXED_POINT );
			int y2 = cy + ( ( x   *s   +(y+h)*c-c2)>>FIXED_POINT );
			int x3 = cx + ( ((x+w)*c-c2-(y+h)*s+s2)>>FIXED_POINT );
			int y3 = cy + ( ((x+w)*s-s2+(y+h)*c-c2)>>FIXED_POINT );
			int x4 = cx + ( ((x+w)*c-c2- y   *s   )>>FIXED_POINT );
			int y4 = cy + ( ((x+w)*s-s2+ y   *c   )>>FIXED_POINT );
			g.fillTriangle(x1,y1, x2,y2, x3,y3);
			g.fillTriangle(x3,y3, x4,y4, x1,y1);
		}
	}
}
